/*
 * Author: Jake Goodwin
 * Date: 2023
 * Description: Small library to communicate with the AT-09 bluetooth module. 
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <stdint.h>
#include <util/delay.h>
#include "avr_usart.h"


//#############################
//Globals
//#############################

int main() {
 
    led_blink();
    led_blink(); 
    led_blink(); 
    _delay_ms(1000);
    _delay_ms(1000);

    init_usart0(); 

    //enable interrupts
    sei();
    
    uint8_t buf_len = 12;
    //unsigned char data[12] = "snd 12chars\0";
    //unsigned char data[1] = "A";
    unsigned char data[1] = "B"; 

    while(1) {
        //serial0_write(data, buf_len);
        if(data[0] == 0) {
            data[0] = '!';
        }
        tx_usart0(data[0]);
        //serial0_read(data, buf_len);
        //data = rx_usart0();
        serial0_flush_rxbuf();
        serial0_read_with_err_checking(data, 1);
    }
    
    return 0;
}



//#############################
//FUNCTIONS
//#############################

/*
 * Input: None
 * Output: None
 * Description: init usart0 hardware in async mode
 */
void init_usart0(void) {
    //DDRB |= (1<<TX_DDR)|(0<<RX_DDR);
    //PORTD |= (0<<TX_PIN)|(0<<RX_PIN); 
    
    //setup stuff for usart communications.
    
    //set baud rate, this is 9600Baud with a 8Mhz clock
    //UBRR0H |= (uint8_t) (BT_UBRR>>8);
    //UBRR0L |= (uint8_t) BT_UBRR;
    UBRR0H |= (uint8_t) (0x33>>8);
    UBRR0L |= (uint8_t) 0x33;

    //Enable recv and Transmit pins, overrides other uses.
    //IN the usart control and status register 0B
    UCSR0B = (1<<RXEN0) | (1<<TXEN0);

    //setting the data frame format
    //We leave the 2MSB zero for async serial. And we also don't enable
    //parity bits for now. We set a 2bit stop bit and 8bit char size.
    //UCSR0C = (1<<USBS0) | (3<<UCSZ00); 
    UCSR0C = (1<<USBS0) | (1<<UCSZ00) | (1<<UCSZ01) | (0<<UCSZ02);
}


/*
 * Input: a 8bit char
 * Output: None
 * Description: Sends a char over usart0
 */
void tx_usart0(unsigned char data) {
    //first wait for the transmit buffer to be empty
    while(!(UCSR0A & (1<<UDRE0))) {
       ;//Equiv of continue
    }

    //now that it's empty, fill buffer with TX data.
    UDR0 = data;
}



/*
 * Input:None 
 * Output: a 8bit char
 * Description: Recvs a char over usart0
 */
unsigned char rx_usart0(void) {
    //first wait for the data to be received.
    while( !(UCSR0A & (1<<RXC0)) ) {
        ;//Equiv of continue
    }

    //now return the data
    return UDR0;
}


void serial0_flush_rxbuf(void) {
    uint8_t none;
    while( UCSR0A & (1<<RXC0)) {
        none = UDR0;
    }
}


/*
 * Inupt: None
 * Output: None
 * Description: Nothing right now.
 */
void serial0_enable_timeouts(uint8_t ms) {
    ;
}

/*
 * Inupt: None
 * Output: None
 * Description:
 */
void serial0_write(unsigned char* buffer, uint8_t write_length) {
    for(uint8_t i = 0; i < write_length; i++) {
        tx_usart0(buffer[i]);
    }
}

/*
 * Inupt: A serialbuffer and length
 * Output: None
 * Description: Reads the serial data into a buffer of x length.
 */
void serial0_read(unsigned char* buffer, uint8_t buf_length) {
    //clear buffer out?
    //buffer = 0;
    for(uint8_t i = 0; i < buf_length; i++) {
        buffer[i] = rx_usart0();
    }
}


/*
 * Inupt: A serialbuffer and length
 * Output: None
 * Description: Reads the serial data into a buffer of x length, 
 * it does error checking from the register first.
 */
void serial0_read_with_err_checking(unsigned char* buffer, uint8_t buf_length) {
     //first wait for the data to be received.
    while( !(UCSR0A & (1<<RXC0)) ) {
        ;//Equiv of continue
    }
    
    //check for errors in the register.
    //Check for frame error.
    if(UCSR0A & (1<<FE0)) {
        led_blink();
    }
    //check for data over-run
    else if(UCSR0A & (1<<DOR0)) {
        led_blink();
        led_blink();
    }
    //check for parity error
    else if(UCSR0A & (1<<UPE0)) {
        led_blink();
        led_blink();
        led_blink();
    }

    //now return the data
    buffer[0] = UDR0;   
}


/*
 * Inupt: BUAD RATE as uint8_t
 * Output: None
 * Description: Sets the baud rate of the serial usart0
 * VALUES @ 1Mhz:
 * 207 --> 300bps
 * 103 --> 600bps
 * 51 --> 1200bps
 * 25 --> 2400bps
 * 12 --> 4800bps
 * 6 --> 9600bps !!! 7% error rate
 */

void serial0_set_baud(uint8_t baud) {
     UBRR0H |= (uint8_t) (baud>>8);
     UBRR0L |= (uint8_t) baud;    
}


/*
 * Input: None
 * Output: None
 * Description: Toggles the pin for the LED indicator.
 */
static void led_blink(void) {
    //Set the DDR for output.
    DDRC |= (1<<LED_PIN);

    PORTC ^= (1<<LED_PIN);
    _delay_ms(250);
    PORTC ^= (1<<LED_PIN);
    _delay_ms(250);
    
}


